"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var automerge_1 = __importDefault(require("automerge"));
var updateCodeMirrorDocs_1 = __importDefault(require("./updateCodeMirrorDocs"));
var Mutex_1 = __importDefault(require("./Mutex"));
var updateAutomergeDoc_1 = __importDefault(require("./updateAutomergeDoc"));
/**
 * Connect an Automerge document
 *
 * @param watchableDoc - the Automerge document that will be connected to CodeMirror instances
 * @return ConnectCodeMirror - a function for connecting an Automerge.Text object in the document to a CodeMirror instance
 */
function connectAutomergeDoc(watchableDoc) {
    var doc = watchableDoc.get();
    var mutex = new Mutex_1.default();
    var codeMirrorByTextId = new Map();
    function getCodeMirror(textObjectId) {
        return codeMirrorByTextId.get(textObjectId);
    }
    var watchableDocHandler = function () {
        var newDoc = watchableDoc.get();
        updateCodeMirrorDocs_1.default(doc, newDoc, getCodeMirror, mutex);
        doc = newDoc;
    };
    /**
     * Connects a CodeMirror instance to an Automerge.Text object.
     *
     * @param codeMirror the editor instance
     * @param getText a function that looks up the Automerge.Text object to connect the CodeMirror editor to
     */
    function connectCodeMirror(codeMirror, getText) {
        var text = getText(watchableDoc.get());
        if (!text) {
            throw new Error("Cannot connect CodeMirror. Did not find text in " + JSON.stringify(watchableDoc.get()));
        }
        if (codeMirrorByTextId.size === 0) {
            // We only start listening for Automerge document changes when the first CodeMirror instance is connected.
            watchableDoc.registerHandler(watchableDocHandler);
        }
        // Establish the association between the Automerge.Text objectId and the CodeMirror instance.
        // This association is used during the processing of diffs between two Automerge document changes,
        // so the right CodeMirror instance can be found for each Automerge.Text change.
        codeMirrorByTextId.set(automerge_1.default.getObjectId(text), codeMirror);
        codeMirror.setValue(getText(watchableDoc.get()).toString());
        var codeMirrorChangeHandler = function (editor, change) {
            if (change.origin !== 'automerge') {
                mutex.lock();
                var newDoc = updateAutomergeDoc_1.default(watchableDoc.get(), getText, editor.getDoc(), change);
                watchableDoc.set(newDoc);
                mutex.release();
            }
        };
        codeMirror.on('change', codeMirrorChangeHandler);
        function disconnectCodeMirror() {
            codeMirror.off('change', codeMirrorChangeHandler);
            codeMirrorByTextId.delete(automerge_1.default.getObjectId(text));
            if (codeMirrorByTextId.size === 0) {
                // We're no longer connected to any CodeMirror instances, so stop listening for Automerge document changes
                watchableDoc.unregisterHandler(watchableDocHandler);
            }
        }
        return disconnectCodeMirror;
    }
    return connectCodeMirror;
}
exports.default = connectAutomergeDoc;
//# sourceMappingURL=connectAutomergeDoc.js.map